package LinearAlgebra;

import java.io.*;

import java.util.Vector;
import Paintable.*;

public class Curve extends GraphAbstract {
    private Vector parameters;
    private int unitParametrizedUntilIndex;

    protected Curve() {
	super();
	parameters = new Vector(100);
	unitParametrizedUntilIndex = -1;
    }

    public Curve(Sample sample) {
	super(sample);
	parameters = new Vector(GetSize());
	for(int i=0;i<GetSize();i++)
	    parameters.addElement(new Double(0));
	unitParametrizedUntilIndex = -1;
    }

    public Sample Clone() {
	Curve curve = new Curve(super.Clone());
	curve.parameters = new Vector(GetSize());
	for(int i=0;i<unitParametrizedUntilIndex;i++)
	    curve.parameters.setElementAt(parameters.elementAt(i),i);
	curve.unitParametrizedUntilIndex = unitParametrizedUntilIndex;
	return curve;
    }

    public Sample ShallowClone() {
	Curve curve = new Curve(super.ShallowClone());
	curve.parameters = parameters;
	curve.unitParametrizedUntilIndex = unitParametrizedUntilIndex;
	return curve;
    }

    public Sample DefaultClone() {
	return new Curve(new Sample());
    }

    final public void Reset() {
	super.Reset();
	parameters.setSize(0);
	unitParametrizedUntilIndex = -1;
    }

    final public void InsertPointAt(Vektor vektor,int index) {
	super.InsertPointAt(vektor,index);
	parameters.insertElementAt(new Double(0),index);
	if (unitParametrizedUntilIndex >= index)
	    unitParametrizedUntilIndex = index - 1;
	
    } 

    final public void DeletePointAt(int index) {
	super.DeletePointAt(index);
	if (unitParametrizedUntilIndex >= index)
	    unitParametrizedUntilIndex = index - 1;
    }

    final public void UpdatePointAt(Vektor vektor,int index) {
	super.UpdatePointAt(vektor,index);
	if (unitParametrizedUntilIndex >= index)
	    unitParametrizedUntilIndex = index - 1;
    }

    final public void SetPointAt(Vektor vektor,int index) {
	super.SetPointAt(vektor,index);
	if (unitParametrizedUntilIndex >= index)
	    unitParametrizedUntilIndex = index - 1;
    }

    final private void SetParameterAt(double parameter,int index) {
	parameters.setElementAt(new Double(parameter),index);
    }

    final public double GetParameterAt(int index) {
	return ((Double)(parameters.elementAt(index))).doubleValue();
    }

    // Unit-speed (Euclidean distance)
    final public void UnitParametrize() {
	if (unitParametrizedUntilIndex < GetSize() - 1) {
	    double t;
	    if (unitParametrizedUntilIndex == -1)
		t = 0.0;
	    else
		t = GetParameterAt(unitParametrizedUntilIndex);
	    for(int i=unitParametrizedUntilIndex + 1;i<GetSize();i++) {
		SetParameterAt(t,i);
		if (i < GetSize() - 1)
		    t += GetPointAt(i).Dist2(GetPointAt(i+1));
	    }
	    unitParametrizedUntilIndex = GetSize() - 1;
	}
    }    
    
    final public double GetLength() { 
	UnitParametrize();
	if (GetSize() > 0)
	    return GetParameterAt(GetSize() - 1); 
	else
	    return 0;
    }

    final public Vektor GetPointAtParameter(double parameter) { 
	UnitParametrize();
	if (parameter < 0)
	    parameter = 0;
	else if (GetSize() > 0 && parameter > GetParameterAt(GetSize()-1))
	    parameter = GetParameterAt(GetSize()-1);
	for(int i=0;i<GetSize();i++) {
	    if (GetParameterAt(i) == parameter)
		return GetPointAt(i).Clone();
	    else if (GetParameterAt(i) > parameter)
		return GetLineSegmentAt(i - 1).GetPointAtParameter(GetParameterAt(i) - parameter);
	}
	// just for the compiler...
	return GetPointAt(GetSize()-1).Clone();
    }

    final public Vektor GetPointAtRandomParameter() {
	return GetPointAtParameter(GetLength() * Math.random());
    }

    // should go to subclass
//      final public Vektor GetPointAtRandomParameter(Functions.Function f) {
//  	double y0 = f.Phi(0);
//  	double y1 = f.Phi(1);
//  	Functions.Function distribution = 
//  	    new Functions.ScaleYFunction(1/(y1 - y0),new Functions.TranslateYFunction(-y0,f));
//  	return GetPointAtParameter(GetLength() * distribution.Phi(Math.random()));
//      }

    // Overwriting GraphAbstract's Save
    public void Save(String fileName) {
	try {
	    FileOutputStream fOut = new FileOutputStream(fileName);
	    PrintStream pOut = new PrintStream(fOut);
	    for(int i=0;i<GetSize();i++) {
		GetPointAt(i).Save(pOut);
		pOut.println();
	    }
	    pOut.close();
	    fOut.close();
	} catch(IOException e) {
	    System.out.println("Can't open file " + fileName);
	}
    }

    // Overwriting GraphAbstract's SaveToString
    final public String SaveToString() {
	String str = new String();
	for(int i=0;i<GetSize();i++)
	    str += GetPointAt(i).SaveToString() + "\n";
	return str;
    }
    
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                          abstract functions from GraphAbstract BEGIN
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    final public int GetVektorIndex1OfEdgeAt(int index) {
	return index;
    }

    final public int GetVektorIndex2OfEdgeAt(int index) {
	return index + 1;
    }

    final public int GetNumOfLineSegments() {
	return GetSize() - 1;
    }
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                          abstract functions from GraphAbstract END
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                    interface PaintableInterface BEGIN
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Continuously drawing curve instead of drawing line segments as in GraphAbstracts, so dashed lines are nice
    public void PrintToPostScript(DataPostScriptDocument ps,double pixelWidth,String type,String title) {
	if (GetSize() > 1) {
	    if (type.equals("")) {
		ps.GSave();
		if (ps.BlackAndWhite())
		    ps.SetGray(0); //black
		ps.NewPath();
		for(int i=0;i<GetSize();i++) {
		    ps.NewLine();
		    ps.Push(ps.ConvertXCoordinate(GetPointAt(i).GetCoordX(ps.projectionPlane)));
		    ps.Push(ps.ConvertYCoordinate(GetPointAt(i).GetCoordY(ps.projectionPlane)));
		}
		ps.NewLine();
		ps.SetLineWidth((pixelWidth + 1) * ps.GetWidthUnit());
		ps.MoveTo();
		ps.BeginFor(1,1,GetSize());
		{
		    ps.Pop();
		    ps.LineTo();
		}
		ps.EndFor();
		if (!title.equals("")) {
		    ps.MoveTo(ps.GetNextLineLeftX(),ps.GetNextLineCenterY());
		    ps.LineTo(ps.GetNextLineRightX(),ps.GetNextLineCenterY());
		    ps.Stroke();	
		    ps.PrintText(title,ps.GetNextLineTextX(),ps.GetNextLineTextY());
		    ps.NextLine();
		}
		ps.Stroke();	
		ps.GRestore();
	    }
	    else {
		super.PrintToPostScript(ps,pixelWidth,type,title);
	    }
	}
    }
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                    interface PaintableInterface END
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}



